/* YapConfig.cpp: Yap configuration settings

   Copyright (C) 1996-2007 Christian Schenk

   This file is part of Yap.

   Yap is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
   
   Yap is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.
   
   You should have received a copy of the GNU General Public License
   along with Yap; if not, write to the Free Software Foundation, 59
   Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

#include "StdAfx.h"

#include "yap.h"

auto_ptr<YapConfig> g_pYapConfig;

#define DEFAULT_DVIPAGE_MODE DviPageMode::Auto
#define DEFAULT_DVIPAGE_MODE_STRING \
  EnumToString<DviPageMode>(DviPageMode::Auto, dviPageModes)

#define DEFAULT_UNIT Units::BigPoints
#define DEFAULT_UNIT_STRING \
  EnumToString<Units>(Units::BigPoints, units)

template<class Enum>
struct EnumAndString
{
  typename Enum::EnumType num;
  const char * lpsz;
};

namespace
{
  const EnumAndString<DviPageMode> dviPageModes[] = {
    { DviPageMode::Auto, "auto" },
    { DviPageMode::Pk, "pk" },
    { DviPageMode::Dvips, "dvips" },
    { DviPageMode::None, 0 },
  };

  const EnumAndString<Units> units[] = {
    { Units::BigPoints, "bp" },
    { Units::Centimeters, "cm" },
    { Units::Inches, "in" },
    { Units::Millimeters, "mm" },
    { Units::Picas, "pc" },
    { Units::None, 0 },
  };


  const char * const DEFAULT_SECURE_COMMANDS = "\
gunzip,\
zcat,\
bunzip2,\
bzcat";

  const char * const DEFAULT_GAMMA_VALUES = "\
0.01 \
0.05 \
0.1 \
0.2 \
0.3 \
0.4 \
0.5 \
0.6 \
0.7 \
0.8 \
0.9 \
1.0 \
1.25 \
1.5 \
1.75 \
2.0 \
3.0 \
4.0 \
5.0 \
7.5 \
10.0 \
100.0";
};

/* _________________________________________________________________________

   EnumToString
   _________________________________________________________________________ */

template<class Enum>
const char *
EnumToString (/*[in]*/ typename Enum::EnumType		num,
	      /*[in]*/ const EnumAndString<Enum> *	pMap)
{
  for (size_t idx = 0; pMap[idx].lpsz != 0; ++ idx)
    {
      if (pMap[idx].num == num)
	{
	  return (pMap[idx].lpsz);
	}
    }
  UNEXPECTED_CONDITION ("EnumToString");
}

/* _________________________________________________________________________

   StringToEnum
   _________________________________________________________________________ */

template<class Enum>
typename Enum::EnumType
StringToEnum (/*[in]*/ const char *		lpsz,
	      /*[in]*/ const EnumAndString<Enum> *	pMap)
{
  for (size_t idx = 0; pMap[idx].lpsz != 0; ++ idx)
    {
      if (StringCompare(lpsz, pMap[idx].lpsz, true) == 0)
	{
	  return (pMap[idx].num);
	}
    }
  UNEXPECTED_CONDITION ("StringToEnum");
}

/* _________________________________________________________________________

   GetProfileBool
   _________________________________________________________________________ */

bool
GetProfileBool (/*[in]*/ const char *	lpszKeyName,
		/*[in]*/ const char *	lpszValueName,
		/*[in]*/ bool			defaultValue)
{
   CWinApp * pApp = AfxGetApp();
   ASSERT_VALID (pApp);
   return (pApp->GetProfileInt(lpszKeyName,
			       lpszValueName,
			       static_cast<int>(defaultValue))
	   != 0);
}

/* _________________________________________________________________________

   GetProfileBool
   _________________________________________________________________________ */

double
GetProfileDouble (/*[in]*/ const char *	lpszKeyName,
		  /*[in]*/ const char *	lpszValueName,
		  /*[in]*/ double		fDefault)
{
   CWinApp * pApp = AfxGetApp();
   ASSERT_VALID (pApp);
   CString strDefault;
   strDefault.Format ("%f", fDefault);
   CString str =
     pApp->GetProfileString(lpszKeyName, lpszValueName, strDefault);
   double dbl = _tstof(str);
   return (dbl);
}

/* _________________________________________________________________________

   RemoveDefaultValue
   _________________________________________________________________________ */

bool
RemoveDefaultValue (/*[in]*/ const char *	lpszKeyName,
		    /*[in]*/ const char *	lpszValueName)
{
  CString strPath;
  strPath.Format ("Software\\MiK\\MiKTeX\\yap\\%s", lpszKeyName);
  HKEY hKey;
  LONG res =
    RegOpenKeyEx(HKEY_CURRENT_USER,
		 strPath,
		 0,
		 KEY_ALL_ACCESS,
		 &hKey);
  if (res == ERROR_SUCCESS)
    {
      res = RegDeleteValue(hKey, lpszValueName);
      RegCloseKey (hKey);
    }
  return (res == ERROR_SUCCESS);
}

/* _________________________________________________________________________

   UpdateRegValue
   _________________________________________________________________________ */

void
UpdateRegValue (/*[in]*/ const char *	lpszKeyName,
		/*[in]*/ const char *	lpszValueName,
		/*[in]*/ bool			value,
		/*[in]*/ bool			defaultValue)
{
  if (value == defaultValue && RemoveDefaultValue(lpszKeyName, lpszValueName))
    {
      return;
    }
  bool oldValue = GetProfileBool(lpszKeyName, lpszValueName, defaultValue);
  if (oldValue != value)
    {
      CWinApp * pApp = AfxGetApp();
      ASSERT_VALID (pApp);
      pApp->WriteProfileInt (lpszKeyName, lpszValueName, (value ? 1 : 0));
    }
}

/* _________________________________________________________________________

   UpdateRegValue
   _________________________________________________________________________ */

void
UpdateRegValue (/*[in]*/ const char *	lpszKeyName,
		/*[in]*/ const char *	lpszValueName,
		/*[in]*/ double			value,
		/*[in]*/ double			defaultValue)
{
  if (value == defaultValue && RemoveDefaultValue(lpszKeyName, lpszValueName))
    {
      return;
    }
  double oldValue = GetProfileDouble(lpszKeyName, lpszValueName, defaultValue);
  if (oldValue != value)
    {
      CWinApp * pApp = AfxGetApp();
      ASSERT_VALID (pApp);
      CString strVal;
      strVal.Format ("%f", value);
      pApp->WriteProfileString (lpszKeyName, lpszValueName, strVal);
    }
}

/* _________________________________________________________________________

   UpdateRegValue
   _________________________________________________________________________ */

void
UpdateRegValue (/*[in]*/ const char *	lpszKeyName,
		/*[in]*/ const char *	lpszValueName,
		/*[in]*/ int			value,
		/*[in]*/ int			defaultValue)
{
  if (value == defaultValue && RemoveDefaultValue(lpszKeyName, lpszValueName))
    {
      return;
    }
  CWinApp * pApp = AfxGetApp();
  ASSERT_VALID (pApp);
  int oldValue = pApp->GetProfileInt(lpszKeyName, lpszValueName, defaultValue);
  if (oldValue != value)
    {
      pApp->WriteProfileInt (lpszKeyName, lpszValueName, value);
    }
}

/* _________________________________________________________________________

   UpdateRegValue
   _________________________________________________________________________ */

void
UpdateRegValue (/*[in]*/ const char *		lpszKeyName,
		/*[in]*/ const char *		lpszValueName,
		/*[in]*/ const CString &		value,
		/*[in]*/ const char *		lpszDefaultValue)
{
  if (value == (lpszDefaultValue ? lpszDefaultValue : "")
      && RemoveDefaultValue(lpszKeyName, lpszValueName))
    {
      return;
    }
  CWinApp * pApp = AfxGetApp();
  ASSERT_VALID (pApp);
  CString oldValue
    = pApp->GetProfileString(lpszKeyName, lpszValueName, lpszDefaultValue);
  if (! (oldValue == value))
    {
      pApp->WriteProfileString (lpszKeyName, lpszValueName, value);
    }
}

/* _________________________________________________________________________

   SetMetafontMode
   _________________________________________________________________________ */

bool
SetMetafontMode (/*[in]*/ const char *	lpszMnemonic,
		 /*[out]*/ unsigned long &	metafontModeIdx)
{
  MIKTEXMFMODE metafontMode;
  for (unsigned long idx = 0;
       SessionWrapper(true)->GetMETAFONTMode(idx, &metafontMode);
       ++ idx)
    {
      if (StringCompare(lpszMnemonic, metafontMode.szMnemonic) == 0)
	{
	  metafontModeIdx = idx;
	  return (true);
	}
    }
  return (false);
}

/* _________________________________________________________________________

   YapConfig::SetDisplayMetafontMode
   _________________________________________________________________________ */

bool
YapConfig::SetDisplayMetafontMode (/*[in]*/ const char * lpszMnemonic)
{
  return (SetMetafontMode(lpszMnemonic, displayMetafontMode));
}

/* _________________________________________________________________________

   YapConfig::SetPrinterMetafontMode
   _________________________________________________________________________ */

bool
YapConfig::SetPrinterMetafontMode (/*[in]*/ const char * lpszMnemonic)
{
  return (SetMetafontMode(lpszMnemonic, printerMetafontMode));
}

/* _________________________________________________________________________

   YapConfig::Load
   _________________________________________________________________________ */

void
YapConfig::Load ()
{
  CWinApp * pApp = AfxGetApp();

  ASSERT_VALID (pApp);

  //
  // HKCU/Software/MiK/MiKTeX/Yap/Settings
  //
  checkFileTypeAssociations =
    GetProfileBool("Settings",
		   "Check Associations",
		   true);
  inverseSearchCommandLine =
    pApp->GetProfileString("Settings",
			   "Editor",
			   "notepad \"%f\"");
  lastTool =
    pApp->GetProfileInt("Settings",
			"Last Tool",
			0);
  maintainHorizontalPosition =
    GetProfileBool("Settings",
		   "Maintain Horizontal Position",
		   false);
  maintainVerticalPosition =
    GetProfileBool("Settings",
		   "Maintain Vertical Position",
		   false);

#if defined(MIKTEX_DEBUG)
  showBoundingBoxes =
    GetProfileBool("Settings",
		   "Show Bounding Boxes",
		   false);
#endif
  showSourceSpecials =
    GetProfileBool("Settings",
		   "Show Source Specials",
		   false);

  //
  // HKCU/Software/MiK/MiKTeX/Yap/screen
  //
  continuousView =
    GetProfileBool("screen",
		   "ContinuousView",
		   true);
#if WIN95NOCONTVIEW
  if (continuousView && win95)
    {
      continuousView = false;
    }
#endif
  doublePage =
    GetProfileBool("screen",
		   "doublepage",
		   false);
  gamma =
    GetProfileDouble("screen",
		     "gamma",
		     1.0);
  if (gamma <= 0.0)
    {
      gamma = 1.0;
    }
  gammaValues =
    pApp->GetProfileString("screen",
			   "gammavalues",
			   DEFAULT_GAMMA_VALUES);
  CString displayMode =
    pApp->GetProfileString("screen",
			   "mode",
			   "ljfour");
  if (! SetDisplayMetafontMode(displayMode))
    {
      FATAL_MIKTEX_ERROR ("YapConfig::Load",
			  T_("Unknown METAFONT mode."),
			  static_cast<const char *>(displayMode));
    }
  displayShrinkFactor =
    pApp->GetProfileInt("screen",
			"shrinkfactor",
			6);
  if (displayShrinkFactor == 0)
    {
      displayShrinkFactor = 1;
    }
  renderGraphicsInBackground =
    GetProfileBool("screen",
		   "graphicsinbackground",
		   DEFAULT_bRenderGraphicsInBackground);
  dviPageMode =
    StringToEnum<DviPageMode>
    (pApp->GetProfileString("screen",
			    "dvipagemode",
			    DEFAULT_DVIPAGE_MODE_STRING),
     dviPageModes);
  unit =
    StringToEnum<Units>
    (pApp->GetProfileString("screen",
			    "unit",
			    DEFAULT_UNIT_STRING),
     units);

  //
  // HKCU/Software/MiK/MiKTeX/Yap/printer
  //
  CString printerMode =
    pApp->GetProfileString("printer",
			   "mode",
			   "ljfour");
  if (! SetPrinterMetafontMode(printerMode))
    {
      FATAL_MIKTEX_ERROR ("YapConfig::Load",
			  T_("Unknown METAFONT mode."),
			  static_cast<const char *>(printerMode));
    }
  pageXShift =
    pApp->GetProfileInt("printer",
			"nPageXShift",
			0);
  pageYShift =
    pApp->GetProfileInt("printer",
			"nPageYShift",
			0);

  //
  // HKCU/Software/MiK/MiKTeX/Yap/DVIPS
  //
  dvipsExtraOptions =
    pApp->GetProfileString("DVIPS",
			   "Extra Options");

  //
  // HKCU/Software/MiK/MiKTeX/Yap/Magnifying Glass
  //
  magGlassHidesCursor =
    GetProfileBool("Magnifying Glass",
		   "Hide Cursor",
		   false);
  magGlassOpacity =
    pApp->GetProfileInt("Magnifying Glass",
			"Opacity",
			70);

  //
  // HKCU/Software/MiK/MiKTeX/Yap/Small Magnifying Glass
  //
  magGlassSmallHeight =
    pApp->GetProfileInt("Small Magnifying Glass",
			"Height",
			150);
  magGlassSmallShrinkFactor =
    pApp->GetProfileInt("Small Magnifying Glass",
			"Shrink Factor",
			1);
  if (magGlassSmallShrinkFactor == 0)
    {
      magGlassSmallShrinkFactor = 1;
    }
  magGlassSmallWidth =
    pApp->GetProfileInt("Small Magnifying Glass",
			"Width",
			200);

  //
  // HKCU/Software/MiK/MiKTeX/Yap/Medium Magnifying Glass
  //
  magGlassMediumHeight =
    pApp->GetProfileInt("Medium Magnifying Glass",
			"Height",
			250);
  magGlassMediumShrinkFactor =
    pApp->GetProfileInt("Medium Magnifying Glass",
			"Shrink Factor",
			1);
  if (magGlassMediumShrinkFactor == 0)
    {
      magGlassMediumShrinkFactor = 1;
    }
  magGlassMediumWidth =
    pApp->GetProfileInt("Medium Magnifying Glass",
			"Width",
			400);

  //
  // HKCU/Software/MiK/MiKTeX/Yap/Large Magnifying Glass
  //
  magGlassLargeHeight =
    pApp->GetProfileInt("Large Magnifying Glass",
			"Height",
			500);
  magGlassLargeShrinkFactor =
    pApp->GetProfileInt("Large Magnifying Glass",
			"Shrink Factor",
			1);
  if (magGlassLargeShrinkFactor == 0)
    {
      magGlassLargeShrinkFactor = 1;
    }
  magGlassLargeWidth =
    pApp->GetProfileInt("Large Magnifying Glass",
			"Width",
			700);

  //
  // HKCU/Software/MiK/MiKTeX/Yap/Security
  //
  enableShellCommands =
    pApp->GetProfileInt("Security",
			"EnableEmbeddedCommands",
			SEC_ASK_USER);
  secureCommands =
    pApp->GetProfileString("Security",
			   "SecureCommands",
			   DEFAULT_SECURE_COMMANDS);
}

/* _________________________________________________________________________

   YapConfig::Save
   _________________________________________________________________________ */

void
YapConfig::Save ()
{
  MIKTEXMFMODE mfmode;

  //
  // HKCU/Software/MiK/MiKTeX/Yap/Settings
  //
  UpdateRegValue ("Settings",
		  "Check Associations",
		  checkFileTypeAssociations,
		  true);
  UpdateRegValue ("Settings",
		  "Editor",
		  inverseSearchCommandLine,
		  "notepad \"%f\"");
  UpdateRegValue ("Settings",
		  "Last Tool",
		  lastTool,
		  0);
  UpdateRegValue ("Settings",
		  "Maintain Horizontal Position",
		  maintainHorizontalPosition,
		  false);
  UpdateRegValue ("Settings",
		  "Maintain Vertical Position",
		  maintainVerticalPosition,
		  false);
#if defined(MIKTEX_DEBUG)
  UpdateRegValue ("Settings",
		  "Show Bounding Boxes",
		  showBoundingBoxes,
		  false);
#endif
  UpdateRegValue ("Settings",
		  "Show Source Specials",
		  showSourceSpecials,
		  false);

  //
  // HKCU/Software/MiK/MiKTeX/Yap/screen
  //
  UpdateRegValue ("screen",
		  "ContinuousView",
		  continuousView,
		  true);
  UpdateRegValue ("screen",
		  "DoublePage",
		  doublePage,
		  false);
  UpdateRegValue ("screen",
		  "gamma",
		  gamma,
		  1.0);
  UpdateRegValue ("screen",
		  "gammavalues",
		  gammaValues,
		  DEFAULT_GAMMA_VALUES);
  if (SessionWrapper(true)->GetMETAFONTMode(displayMetafontMode, &mfmode))
    {
      UpdateRegValue ("screen",
		      "mode",
		      CString(mfmode.szMnemonic),
		      "ljfour");
    }
  UpdateRegValue ("screen",
		  "shrinkfactor",
		  displayShrinkFactor,
		  6);
  UpdateRegValue ("screen",
		  "graphicsinbackground",
		  renderGraphicsInBackground,
		  DEFAULT_bRenderGraphicsInBackground);
  UpdateRegValue ("screen",
		  "dvipagemode",
		  CString(EnumToString<DviPageMode>(dviPageMode.Get(),
						    dviPageModes)),
		  DEFAULT_DVIPAGE_MODE_STRING);
  UpdateRegValue ("screen",
		  "unit",
		  CString(EnumToString<Units>(unit.Get(),
					      units)),
		  DEFAULT_UNIT_STRING);

  //
  // HKCU/Software/MiK/MiKTeX/Yap/printer
  //
  if (SessionWrapper(true)->GetMETAFONTMode(printerMetafontMode, &mfmode))
    {
      UpdateRegValue ("printer",
		      "mode",
		      CString(mfmode.szMnemonic),
		      "ljfour");
    }
  UpdateRegValue ("printer",
		  "nPageXShift",
		  pageXShift,
		  0);
  UpdateRegValue ("printer",
		  "nPageYShift",
		  pageYShift,
		  0);

  //
  // HKCU/Software/MiK/MiKTeX/Yap/DVIPS
  //
  UpdateRegValue ("DVIPS",
		  "Extra Options",
		  dvipsExtraOptions,
		  static_cast<const char *>(0));

  //
  // HKCU/Software/MiK/MiKTeX/Yap/Magnifying Glass
  //
  UpdateRegValue ("Magnifying Glass",
		  "Hide Cursor",
		  magGlassHidesCursor,
		  false);
  UpdateRegValue ("Magnifying Glass",
		  "Opacity",
		  magGlassOpacity,
		  70);

  //
  // HKCU/Software/MiK/MiKTeX/Yap/Small Magnifying Glass
  //
  UpdateRegValue ("Small Magnifying Glass",
		  "Height",
		  magGlassSmallHeight,
		  150);
  UpdateRegValue ("Small Magnifying Glass",
		  "Shrink Factor",
		  magGlassSmallShrinkFactor,
		  1);
  UpdateRegValue ("Small Magnifying Glass",
		  "Width",
		  magGlassSmallWidth,
		  200);

  //
  // HKCU/Software/MiK/MiKTeX/Yap/Medium Magnifying Glass
  //
  UpdateRegValue ("Medium Magnifying Glass",
		  "Height",
		  magGlassMediumHeight,
		  250);
  UpdateRegValue ("Medium Magnifying Glass",
		  "Shrink Factor",
		  magGlassMediumShrinkFactor,
		  1);
  UpdateRegValue ("Medium Magnifying Glass",
		  "Width",
		  magGlassMediumWidth,
		  400);

  //
  // HKCU/Software/MiK/MiKTeX/Yap/Large Magnifying Glass
  //
  UpdateRegValue ("Large Magnifying Glass",
		  "Height",
		  magGlassLargeHeight,
		  500);
  UpdateRegValue ("Large Magnifying Glass",
		  "Shrink Factor",
		  magGlassLargeShrinkFactor,
		  1);
  UpdateRegValue ("Large Magnifying Glass",
		  "Width",
		  magGlassLargeWidth,
		  700);

  //
  // HKCU/Software/MiK/MiKTeX/Yap/Security
  //
  UpdateRegValue ("Security",
		  "EnableEmbeddedCommands",
		  enableShellCommands,
		  SEC_ASK_USER);
  UpdateRegValue ("Security",
		  "SecureCommands",
		  secureCommands,
		  DEFAULT_SECURE_COMMANDS);
}

/* _________________________________________________________________________

   YapConfig::YapConfig
   _________________________________________________________________________ */

YapConfig::YapConfig ()
  : dviPageMode (DEFAULT_DVIPAGE_MODE),
    unit (Units::None),
    win95 ((GetVersion() & 0x80000000) != 0)
{
  Load ();
}

/* _________________________________________________________________________

   YapConfig::~YapConfig
  _________________________________________________________________________ */

YapConfig::~YapConfig ()
{
#if 0
  try
    {
      Save ();
    }
  catch (const exception &)
    {
    }
#endif
}
